from .AsyncScraper import KunYu
from .SyncScraper import SyncKunYu

__all__ = ["KunYu", "SyncKunYu"]

# Package metadata
__version__ = "1.1.4"
__author__ = "Muhammad MuQiT"
__email__ = "antasata678@gmail.com"
__license__ = "GPL-V3.0"

